//Glove MIDI Controller
//Cornell ECE 4760 2010
//Anson Dorsey (ajd53), Eric Gunther (ecg35), Jonathon Smythe (jps93)

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <math.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <string.h>

#include <util/delay.h>
#include "uart.h"

#define bufferSize 32

void initialize(void);
void main(void);
void getSensor(void);
void createMIDImessage(char message, char note, char param);

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//software timer
volatile unsigned int time1 = 1;

//keep track of the current sensor being sampled and its value
unsigned char currentSensor = 0;
unsigned char sensorValue = 0;

//8 byte arrays, 1 slot for each sensor
unsigned char sensorCt[8];			//count how long voltage has passed threshold for each sensor
unsigned char sensorMax[8];			//max value for a push - translates to velocity
unsigned char noteOn[8];			//whether or not a note is currently on
unsigned char noteMap[8];			//maps a MIDI note to each sensor
unsigned char noteWait[8];			//whether or not a sensor needs to wait before accepting another press
unsigned char countNoteWait[8];		//counter to wait for another press

unsigned char MIDI_buffer[bufferSize];		//the buffer
volatile unsigned char buffer_bit = 9;		//for sending individual bits in a byte
volatile unsigned char buffer_empty = 1;	
volatile unsigned char *output_ptr, *write_ptr;		//pointers for ring buffer
volatile unsigned char output_tracker = 0;			//tracks for wraparound
volatile unsigned char write_tracker = 0;			//tracks for wraparound

unsigned char mode = 0;			//0 for MIDI, 1 for MATLAB menu
unsigned char sustain = 0;		//0 for no sustain, 1 for sustain

//timer 0 compare ISR
//creates tiemr for sampling sensors
//triggers every .125 ms
ISR (TIMER0_COMPA_vect)
{      
  
  if (time1>0)	--time1;
	
}

//timer2 MIDI out ISR
//clock/send MIDI messages
//MIDI clocks at 31.25 Kbaud (+/- 1%).
ISR (TIMER2_COMPA_vect) 
{

	//MIDI idles at logic 1
	//send out bits
	if ( !buffer_empty ) {

		//first bit is the start bit, logic 0
		if (buffer_bit == 9) {
			
			PORTC &= 0xfe; 
			buffer_bit--;
		
		}

		//send data bits, LEAST SIGNIFICANT BIT FIRST
		else if (buffer_bit > 0) { 
			
			PORTC = (PORTC & 0xfe) | (0x01 & *output_ptr);
			buffer_bit--;
			//shift the byte to be sent to make next bit in LSB slot
			*output_ptr = *output_ptr >> 1;
		
		}
		
		else {
			
			PORTC |= 0x01; //stop bit, logic '1'
			buffer_bit = 9; 

			//if queue is empty set the flag
			if (    (output_tracker == (write_tracker - 1)) 
				 || (output_tracker == 31 && write_tracker == 0) ) 
				 { buffer_empty = 1; }

			//update pointers for next transmission
			if (output_tracker < (bufferSize-1)) {
				//go to next byte in memory
				output_tracker++;
				output_ptr++;
			}
			else {
				//wrap ring back to start
				output_tracker = 0;
				output_ptr = output_ptr - (bufferSize-1);
			}
		}

	}
	//else buffer is empty
	else { PORTC |= 0x01; } //MIDI idles at logic '1'

}

//MAIN
void main(void) {

       //initialize everything
       initialize();

       while(1) {

		   //MATLAB interface mode
		   if(mode == 1)  {
			   
				if(time1 == 0) {
				time1 = 40; //works out to every 10 ms
				   
				//scan for note maps and sustain mode
				scanf("%d%d%d%d%d%d%d%d%d",&noteMap[0],&noteMap[1],&noteMap[2],&noteMap[3],&noteMap[4],&noteMap[5],&noteMap[6],&noteMap[7],&sustain);

				mode = 0;

			   }
			   
		   } else {
		   
			//MIDI playing mode
			if (time1 == 0) {
				time1 = 1;
				
				//SW7 to B0 to change mode
				if(PINB < 255) {
					mode = 1;
				}

				//get sensor value
				getSensor();

				//if sensor value is above threshold
				if(sensorValue > 5) {					
					
					//start counting how long its been above threshold and remember max value
					sensorCt[currentSensor]++;
					if (sensorMax[currentSensor] < sensorValue) 
						sensorMax[currentSensor] = sensorValue;
						
				} else {

					//reset the counter and max
					sensorCt[currentSensor] = 0;
					sensorMax[currentSensor] = 0;
				
				}

				
				//if the sensor value has been above threshold for 10 ms and is not supposed to wait for a press
				if(sensorCt[currentSensor] == 10 && noteWait[currentSensor] == 0) {
	
					//scale velocity to be between 0 and 127 
					if(noteOn[currentSensor] == 0 || sustain == 0) {
						
						//send a note on message
						createMIDImessage(0x90,noteMap[currentSensor],sensorMax[currentSensor]/2);
						noteOn[currentSensor] = 0;
					
					} else {
					
						//send a note off message
						createMIDImessage(0x80,noteMap[currentSensor],0);
						noteOn[currentSensor] = 0;
					
					}

					//reset sensorCt and start waiting to accept another press
					sensorCt[currentSensor] = 0;
					countNoteWait[currentSensor] = 1;

				}

				//if sensor is supposed to be waiting increment how long its been waiting
				if(countNoteWait[currentSensor] == 1) {
					noteWait[currentSensor]++;
				}

				//if waiting for 50 ms time to begin accepting presses again
				if(noteWait[currentSensor] > 50) {

					noteWait[currentSensor] = 0;
					countNoteWait[currentSensor] = 0;
					sensorCt[currentSensor] = 0;
					sensorMax[currentSensor] = 0;
				
				}
				
				//incremement current sensor
				currentSensor++;
	            if (currentSensor > 7) 
					currentSensor = 0;
			
				}

			
		   }
       }

}

//get the sensor value according to the current sensor number
//all global variables
void getSensor(void) {

       //get the sample
       //assign sample to global variable
       //increment sensor index global variable
       //start next conversion

       //get the sample
       sensorValue = ADCH;

		//switch to next Sensor
	   if(currentSensor < 7)
       		ADMUX = (0x20) + currentSensor+1;
	   else
	   		ADMUX = (0x20);

       //start another conversion
       ADCSRA |= (1<<ADSC);

}

//create the MIDI message
//ie map sensor values to notes and decide what to send
void createMIDImessage(char message, char note, char param) {

	//put 3 bytes in the buffer and update the write tracker
	//remember to wrap around if the tracker goes above the buffer size
	
	MIDI_buffer[write_tracker] = message;

	write_tracker++;
	write_ptr++;
	if(write_tracker > 31) {
		write_tracker = 0;
		write_ptr = write_ptr - (bufferSize-1);
	}

	MIDI_buffer[write_tracker] = note;

	write_tracker++;
	write_ptr++;
	if(write_tracker > 31) {
		write_tracker = 0;
		write_ptr = write_ptr - (bufferSize-1);
	}

	MIDI_buffer[write_tracker] = param; 

	write_tracker++;
	write_ptr++;
	if(write_tracker > 31) {
		write_tracker = 0;
		write_ptr = write_ptr - (bufferSize-1);
	}
	
	buffer_empty = 0;
	
}


void initialize(void) {
	//init sensor notes, one octave major scale from C4
	noteMap[0] = 0x3C; //left ring middle c (C4)
	noteMap[1] = 0x3E;
	noteMap[2] = 0x40;
	noteMap[3] = 0x41;
	noteMap[4] = 0x43;
	noteMap[5] = 0x45;
	noteMap[6] = 0x47;
	noteMap[7] = 0x48; // right ring C5

	//initialize ring buffer read and write byte
	output_ptr = &MIDI_buffer[0]; 
	write_ptr = &MIDI_buffer[0];

	//uart for debugging
	uart_init();
	stdout = stdin = stderr = &uart_str;
	//fprintf(stdout,"Init\n\r");

	//for LEDs
	//DDRB=0xff;    // PORT B is an output  
  	//PORTB=0;

	TCCR2B = 0x02; //set to prescale by 8
	OCR2A = 63; //16MHz divided by 8, 64 ticks will trigger ISR at 31.25kHz
	TCCR2A = (1<<WGM21); //set clear on match
	TIMSK2 = (1<<OCIE2A); //initialize interrupt

	DDRC = 0xff;
	PORTC = 0x01; //initialize pin C0 to output for MIDI bits

	//set up timer 0 for .125 mSec timebase 
  	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
    OCR0A = 255;  		//set the compare re to 250 time ticks
	//set prescalar to divide by 8 
    TCCR0B= 2; //0b00000010;	
	// turn on clear-on-match
    TCCR0A= (1<<WGM01);

	//init the A to D converter
 	//channel zero/ left adj /EXTERNAL Aref
 	//!!!CONNECT Aref jumper!!!!
 	ADMUX = 0x20;

   //enable ADC and set prescaler to 1/32* 16MHz=500kHz
   //and clear interupt enable
   //and start a conversion
    ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 5;

	DDRB = 0x00;
	PORTB = 0xff;

	//go interrupts go!
	sei();	
}
